package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLogImpl;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Proxy;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.states.JobStateFactory;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.InputStreamToLogsThread;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.highlevel.util.StringMemorySourceFile;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.rspec.model.LoginService;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.highlevel.util.ProxyServiceUtil;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.dropwizard.jackson.Jackson;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.ConnectionException;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.sftp.SFTPClient;
import net.schmizz.sshj.transport.TransportException;
import org.jetbrains.annotations.Contract;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 17:58
 */
public class TestLinksJob extends JobWithSshConnectionManager<Boolean> {
    private static final Logger LOG = LoggerFactory.getLogger(TestLinksJob.class);
    private static final ObjectMapper MAPPER = Jackson.newObjectMapper();

    private final JFedPreferences jFedPreferences;
    private final ProxyPreferencesManager proxyPreferencesManager;
    private final ProxyServiceUtil proxyServiceUtil;

    private final List<FXRspecNode> failedNodes = Collections.synchronizedList(new ArrayList<>());

    public TestLinksJob(@Nonnull Experiment experiment,
                        @Nonnull HighLevelTaskFactory hltf,
                        @Nonnull TaskThread tt,
                        @Nonnull GeniUserProvider geniUserProvider,
                        @Nonnull JFedPreferences jFedPreferences,
                        @Nonnull ProxyPreferencesManager proxyPreferencesManager,
                        @Nonnull ProxyServiceUtil proxyServiceUtil,
                        @Nonnull JobStateFactory jobStateFactory,
                        @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider) {
        super("Testing Links between Resources", experiment, hltf, tt,
                jobStateFactory, proxySocketFactoryProvider, geniUserProvider);

        this.jFedPreferences = jFedPreferences;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.proxyServiceUtil = proxyServiceUtil;
    }


    @Override
    public Boolean execute() throws Exception {
        LOG.trace("Starting connectivity tests to existing resources in experiment {}", experiment.getName());

        final TestLinksExperimentPartsState tcs = new TestLinksExperimentPartsState();
        setAndRunSshUsingState(tcs);
        closeAllSshConnections();
        return tcs.getStatus() == ExperimentTaskStatus.SUCCESS;
    }

    public class TestLinksExperimentPartsState extends SlicedState<TestLinksExperimentPartStateSlice> {

        private final List<TestLinksExperimentPartStateSlice> slices;

        protected TestLinksExperimentPartsState() {
            super("Test Links of experiment parts");

            slices = experiment.getPartsStream()
                    .filter(part -> part instanceof SfaExperimentPart) //only SFA parts are accessible for the moment
                    .map(TestLinksExperimentPartStateSlice::new)
                    .collect(Collectors.toList());
        }

        @Override
        public Collection<TestLinksExperimentPartStateSlice> getSlices() {
            return slices;
        }

        public List<FXRspecNode> getFailedNodes() {
            return failedNodes;
        }
    }


    public class TestLinksExperimentPartStateSlice extends ExperimentPartStateSlice {
        private TestLinksExperimentPartStateSlice(ExperimentPart experimentPart) {
            super(TestLinksJob.this, experimentPart);
        }

        @Override
        public String getName() {
            return "Testing connectivity to resources in " + experimentPart.getName();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            Collection<FXRspecNode> nodesToTest = findTestableResources(experimentPart);

            if (!nodesToTest.isEmpty()) {
                final TestLinksOfExperimentPartState tcs =
                        new TestLinksOfExperimentPartState(experimentPart, nodesToTest);
                setAndRunState(tcs);

                Map<String, FXRspecNode> failedNodesByClientId =
                        tcs.linkReports.stream()
                                .filter(c -> c.getSummary() != LinkTestResult.SUCCESS)
                                .map(LinkTestReport::getFrom)
                                .collect(Collectors.toMap(FXRspecNode::getClientId, Function.identity(), (a, b)->a));
                failedNodes.addAll(failedNodesByClientId.values());

                return tcs.getStatus();
            } else {
                LOG.debug("Skipping: no testable resources found");
                return ExperimentTaskStatus.SUCCESS;
            }
        }
    }

    @Nonnull
    public Collection<FXRspecNode> findTestableResources(@Nonnull ExperimentPart experimentPart) {
        //use a map to make sure that only one rspecnode per id is collected
        Map<String, FXRspecNode> nodesOfAuthorityWithLoginService = new HashMap<>();
        Set<String> nodeIdsWithoutLoginService = new HashSet<>();

        if (!(experimentPart instanceof SfaExperimentPart)) {
            LOG.debug("Currently no support for connectivity testing of {}", experimentPart.getClass().getName());
            return Collections.emptyList();
        }

        SfaExperimentPart sfaExperimentPart = (SfaExperimentPart) experimentPart;
        for (Sliver sliver : sfaExperimentPart.getSlivers()) {
            //skip fake slivers
            if(sliver.isFake())
                continue;

            //skip virtual slivers (ex. vlan-slivers on ExoGENI)
            if(sliver.isVirtual())
                continue;

            //check if sliver is ready
            if (sliver.getStatus().getGlobalStatus() != StatusDetails.SliverStatus.READY) {
                if (LOG.isTraceEnabled())
                    LOG.trace("Cannot start connectivity test to {} as sliver {} is not READY", experimentPart.getName(),
                            sliver.getUrn());

                //return Collections.emptyList();
                continue;
            }

            //check if manifest is available
            if (sliver.getManifestRspec() == null) {
                if (LOG.isTraceEnabled())
                    LOG.trace("Cannot start connectivity test to {} as manifest is not available for sliver {}",
                            experimentPart.getName(), sliver.getUrn());

//                return Collections.emptyList();
                continue;
            } else {
                //only add node if it has a loginService, and hasn't been added yet to the map
                FXModelRspec modelRspec = sliver.getManifestRspec().getModelRspec(ModelRspecType.FX);
                assert modelRspec != null;
                //noinspection ConstantConditions
                Collection<FXRspecNode> nodes = modelRspec.getNodesByAuthority(sfaExperimentPart.getConnectSfaAuthority());
                for (FXRspecNode node : nodes) {
                    if (!node.getLoginServices().isEmpty()) {
                        if (!nodesOfAuthorityWithLoginService.containsKey(node.getUniqueId())) {
                            nodesOfAuthorityWithLoginService.put(node.getUniqueId(), node);
                            nodeIdsWithoutLoginService.remove(node.getUniqueId()); //remove it from the nodeIdsWithoutLoginService
                        }
                    } else {
                        //check if we haven't found a loginService for this node before, and add it to nodeIdsWithoutLoginService otherwise
                        if (!nodesOfAuthorityWithLoginService.containsKey(node.getUniqueId()))
                            nodeIdsWithoutLoginService.add(node.getUniqueId());
                    }
                }
            }
        }

        if (!nodesOfAuthorityWithLoginService.isEmpty()) {
            return nodesOfAuthorityWithLoginService.values();
        } else if (!nodeIdsWithoutLoginService.isEmpty()) {
            if (LOG.isTraceEnabled())
                LOG.trace("Cannot start connectivity test for authority {} as no nodes with loginServices are available",
                        experimentPart.getName());
            return Collections.emptyList(); //we didn't find a loginService
        } else {
            //start the connectivity test
            if (LOG.isTraceEnabled())
                LOG.trace("No connectivity test will be performed for authority {}, as it isn't responsible for any nodes",
                        experimentPart.getName());
            return Collections.emptyList();   //the authority isn't responsible for a node
        }
    }

    public enum LinkTestResult {
        SUCCESS, PARTIAL_FAILURE, FAILURE;
        boolean isWorseThan(@Nonnull LinkTestResult other) {
            switch (this) {
                case SUCCESS: return false;
                case PARTIAL_FAILURE: return other == SUCCESS;
                case FAILURE: return other != FAILURE;
                default: throw new RuntimeException("Unhandled LinkTestResult"+other);
            }
        }
    }
    public static class LinkTestReport {
        @Nonnull private final FXRspecNode from;
        @Nonnull private final FXRspecNode to;

        private final LinkTestResult summary;

        private final int expectedLinkSpeed_mbps;
        private final String ifaceDevName;
        private final String debug;

        private final boolean connectivityOk;
        private final boolean linkSpeedSettingOk;
        private final boolean throughputOk;
        private final boolean lossOk;

        private final int actualLinkSpeedSetting_mbps;
        private final double actualThroughput_mbps;
        private final double actualLoss_percent;

        public LinkTestReport(@Nonnull FXRspecNode from, @Nonnull FXRspecNode to,
                              int expectedLinkSpeed_mbps, boolean connectivityOk,
                              double actualThroughput_mbps, int actualLinkSpeedSetting_mbps,
                              double actualLoss_percent, String ifaceDevName,
                              String debug) {
            this.from = from;
            this.to = to;
            this.expectedLinkSpeed_mbps = expectedLinkSpeed_mbps;
            this.connectivityOk = connectivityOk;
            this.actualThroughput_mbps = actualThroughput_mbps;
            this.actualLinkSpeedSetting_mbps = actualLinkSpeedSetting_mbps;
            this.actualLoss_percent = actualLoss_percent;
            this.ifaceDevName = ifaceDevName;
            this.debug = debug;

            this.linkSpeedSettingOk = actualLinkSpeedSetting_mbps == expectedLinkSpeed_mbps;
            this.lossOk = actualLoss_percent <= 0.001;
            //low expectations about actual throughput:
            if (expectedLinkSpeed_mbps == 10)
                this.throughputOk = actualThroughput_mbps > 9.0;
            else if (expectedLinkSpeed_mbps == 100)
                this.throughputOk = actualThroughput_mbps > 80.0;  //seen 95 measured
            else if (expectedLinkSpeed_mbps == 1000)
                this.throughputOk = actualThroughput_mbps > 200.0; //currently, between wall physical nodes, between 400-500 is measured
            else
                this.throughputOk = actualThroughput_mbps > (expectedLinkSpeed_mbps / 5.0);

            LinkTestResult summary = LinkTestResult.FAILURE;
            if (connectivityOk && linkSpeedSettingOk && this.throughputOk && lossOk)
                summary = LinkTestResult.SUCCESS;
            else {
                if (connectivityOk || this.throughputOk || lossOk)
                    summary = LinkTestResult.PARTIAL_FAILURE;
            }
            this.summary = summary;
        }

        @Nonnull
        private static <T> T getFromMapOrDefault(Map m, String key, @Nonnull T defaultValue) {
            if (!m.containsKey(key))
                return defaultValue;
            Object res = m.get(key);
            if (res == null)
                return defaultValue;
            return (T) res;
        }

        public static LinkTestReport createFromJson(@Nonnull FXRspecNode from, @Nonnull FXRspecNode to, int expectedLinkSpeed_mbps,
                                                    @Nonnull Map linkTestResMap, String debug) {
            boolean ping = getFromMapOrDefault(linkTestResMap, "ping", false);
            double measured_bw = getFromMapOrDefault(linkTestResMap, "measured_bw", -1.0);
            int iface_speed = getFromMapOrDefault(linkTestResMap, "iface_speed", -1);
            double loss = getFromMapOrDefault(linkTestResMap, "loss", -1.0);
            String iface_dev = getFromMapOrDefault(linkTestResMap, "iface_dev", "unknown");
            String extraDebug;
            try {
                extraDebug = MAPPER.writerWithDefaultPrettyPrinter().writeValueAsString(linkTestResMap);
            } catch (JsonProcessingException e) {
                extraDebug = "Error converting Map to JSON: "+e.getMessage();
            }
            if (debug == null)
                debug = extraDebug;
            else
                debug += "\n\n"+extraDebug;

            return new LinkTestReport(from, to, expectedLinkSpeed_mbps,
                    ping, measured_bw, iface_speed, loss, iface_dev, debug);
        }

        @Nonnull
        public FXRspecNode getFrom() {
            return from;
        }

        @Nonnull
        public FXRspecNode getTo() {
            return to;
        }

        public String getIfaceDevName() {
            return ifaceDevName;
        }

        public String getDebug() {
            return debug;
        }

        public LinkTestResult getSummary() {
            return summary;
        }

        public int getExpectedLinkSpeed_mbps() {
            return expectedLinkSpeed_mbps;
        }

        public boolean isConnectivityOk() {
            return connectivityOk;
        }

        public boolean isLinkSpeedSettingOk() {
            return linkSpeedSettingOk;
        }

        public boolean isThroughputOk() {
            return throughputOk;
        }

        public boolean isLossOk() {
            return lossOk;
        }

        public int getActualLinkSpeedSetting_mbps() {
            return actualLinkSpeedSetting_mbps;
        }

        public double getActualThroughput_mbps() {
            return actualThroughput_mbps;
        }

        public double getActualLoss_percent() {
            return actualLoss_percent;
        }

        @Nullable
        public static LinkTestReport worst(@Nonnull Collection<LinkTestReport> reports) {
            if (reports.isEmpty()) return null;
            LinkTestReport res = reports.iterator().next();
            for (LinkTestReport cur: reports) {
                if (cur.summary.isWorseThan(res.summary))
                    res = cur;
            }
            return res;
        }

        @Nullable
        @Contract("!null,_ -> !null;_,!null -> !null;null,null -> null")
        public static LinkTestReport worstNonNull(@Nullable LinkTestReport a, @Nullable LinkTestReport b) {
            if (a == null && b == null) return null;
            if (a == null) return b;
            if (b == null) return a;
            return worst(a, b);
        }

        @Nonnull
        public static LinkTestReport worst(@Nonnull LinkTestReport a, @Nonnull LinkTestReport b) {
            if (a.summary.isWorseThan(b.summary))
                return a;
            else
                return b;
        }

        @Override
        public String toString() {
            return "LinkTest "+summary+
                    " from=" + from.getClientId() +
                    " to=" + to.getClientId() +
                    " connectivity " + (connectivityOk ? "OK" : "Fail") +
                    " linkSpeed set " + (linkSpeedSettingOk? "OK" : "Fail") +
                    " throughput " + (throughputOk ? "OK" : "Fail")+
                    " loss " + (lossOk ? "OK" : "Fail")+
                    " expected speed=" + expectedLinkSpeed_mbps +" Mbps"+
                    " iface='" + ifaceDevName + '\'' +
                    " set speed=" + actualLinkSpeedSetting_mbps  +" Mbps"+
                    " measured bw=" + actualThroughput_mbps  +" Mbps"+
                    " loss=" + actualLoss_percent +"%"+
                    (debug == null || debug.isEmpty() ? "": " debug='" + debug + '\'');
        }

        public String toStringNoSuccessDebug() {
            return "LinkTest "+summary+
                    " from=" + from.getClientId() +
                    " to=" + to.getClientId() +
                    " connectivity " + (connectivityOk ? "OK" : "Fail") +
                    " linkSpeed set " + (linkSpeedSettingOk? "OK" : "Fail") +
                    " throughput " + (throughputOk ? "OK" : "Fail")+
                    " loss " + (lossOk ? "OK" : "Fail")+
                    " expected speed=" + expectedLinkSpeed_mbps +" Mbps"+
                    " iface='" + ifaceDevName + '\'' +
                    " set speed=" + actualLinkSpeedSetting_mbps  +" Mbps"+
                    " measured bw=" + actualThroughput_mbps  +" Mbps"+
                    " loss=" + actualLoss_percent +"%"+
                    (summary == LinkTestResult.SUCCESS ? "" : (debug == null || debug.isEmpty() ? "": " debug='" + debug + '\''));
        }
    }

    public class TestLinksOfExperimentPartState extends State {

        private final ExperimentPart experimentPart;
        private final Collection<FXRspecNode> nodesToTest;

        private final List<LinkTestReport> linkReports =
                new ArrayList<>();

        private TestLinksOfExperimentPartState(ExperimentPart experimentPart,
                                               Collection<FXRspecNode> nodesToTest) {
            super(String.format("Testing links of %s", experimentPart.getName()));
            this.experimentPart = experimentPart;
            this.nodesToTest = nodesToTest;
        }

        @Nonnull
        @Override
        protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
            assert (!nodesToTest.isEmpty());
            assert experimentPart instanceof SfaExperimentPart;
            assert ((SfaExperimentPart) experimentPart).getSlivers() != null;

            for (Sliver sliver : ((SfaExperimentPart) experimentPart).getSlivers()) {
                assert (sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.READY);
                assert (sliver.isFake() || sliver.getManifestRspec() != null);
            }


            JFedConnection.ProxyInfo proxyInfo = null;
            if (Objects.equals(jFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH), ProxyPreferencesManager.ALWAYS_USE_PROXY)) {
                GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
                assert geniUser.getUserAuthorityServer() != null;
                //noinspection ConstantConditions
                List<Proxy> userAuthProxyInfos = geniUser.getUserAuthorityServer().getTestbed().getProxies();
                proxyInfo =
                        proxyPreferencesManager.getSshProxySettings(userAuthProxyInfos, geniUser);

                if (proxyInfo == null) {
                    LOG.warn("Link tests to nodes on authority {} are skipped as SSH-proxy is active but proxy info was not found." +
                            " (this is a jFed bug)", experimentPart.getName());


                    updateMessage("Skipped because proxy for SSH-connections is active but proxy info was not found. (this is a jFed bug)");
                    experiment.getLinkTestLogger().fireAllReports(linkReports, LinkTestResult.FAILURE);
                    return ExperimentTaskStatus.FAILED;
                }
            }


            LOG.debug("Starting nodes Link test for {} nodes on {}", nodesToTest.size(), experimentPart.getName());

            Iterator<FXRspecNode> it = nodesToTest.iterator();
            while (it.hasNext()) {
                FXRspecNode node = it.next();

                if (node.getLoginServices().isEmpty()) {
                    LOG.debug("Removing node {} from test because it doesn't have any loginServices available", node.getUniqueId());
                    it.remove();
                }
            }

            LinkTestReport allTimeWorst = null;
            for (FXRspecNode node : nodesToTest) {
                List<LinkTestReport> result = null;
                try {
                    result = testLinksOfNode(node, proxyInfo);
                } catch (Exception e) {
                    LOG.debug("Exception while executing link test for node "+node.getUniqueId(), e);
                    updateMessage("Link test failed to run on some nodes.");
                    experiment.getLinkTestLogger().fireAllReports(linkReports, LinkTestResult.FAILURE);
                    return ExperimentTaskStatus.FAILED;
                }

                LinkTestReport curWorst = LinkTestReport.worst(result);
                allTimeWorst = LinkTestReport.worstNonNull(curWorst, allTimeWorst);

                for (LinkTestReport r: result) {
                    LOG.debug("Got LinkReport summary=" + r.getSummary() +
                            " from=" + r.getFrom().getClientId() +
                            " to=" + r.getTo().getClientId());
                }

                linkReports.addAll(result);
//                break; //for debugging
            }

            if (allTimeWorst != null) {
                LOG.debug("allTimeWorst LinkReport summary=" + allTimeWorst.getSummary() +
                        " from=" + allTimeWorst.getFrom().getClientId() +
                        " to=" + allTimeWorst.getTo().getClientId());
            }

            if (allTimeWorst == null || allTimeWorst.getSummary() == LinkTestResult.SUCCESS) {
                experiment.getLinkTestLogger().fireAllReports(linkReports,
                        allTimeWorst == null ? LinkTestResult.SUCCESS : allTimeWorst.summary);
                updateMessage("All links working correctly.");
                return ExperimentTaskStatus.SUCCESS;
            } else {
                experiment.getLinkTestLogger().fireAllReports(linkReports, allTimeWorst.summary);
                LOG.debug("Link Test detected link failures.");
                updateMessage("Link Test detected link failures.");
                //still return success, because we did manage to test it.
                // We succesfully got a result. That result just happens to be a link failure.
                //TODO (The GUI needs to show this better, it's too subtle now. But don't change this return or you break the cli2!)
                return ExperimentTaskStatus.SUCCESS;
//                return ExperimentTaskStatus.FAILED;
            }
        }

        private List<LinkTestReport> testLinksOfNode(FXRspecNode node, JFedConnection.ProxyInfo proxyInfo)
                throws JFedException, InterruptedException, SshException {
            LOG.debug("Testing links from node "+node.getUniqueId());

            LoginService loginService = node.getLoginServices().get(0);

            LOG.trace("Going to try to connect to {} by using loginService {}:{}",
                    node.getUniqueId(), loginService.getHostname(), loginService.getPort());


            JFedConnection.ProxyInfo testbedProxyInfo = proxyServiceUtil.findTestbedProxy(node, loginService);
            if (testbedProxyInfo != null)
                LOG.trace("Trying to make connection to {} with testbed proxy {}", node.getUniqueId(), testbedProxyInfo);

            if (proxyInfo != null && testbedProxyInfo != null) {
                LOG.warn("jFed has proxy currently enabled. However, connectivity test will only use testbed proxy {} while trying to connect to {}!", testbedProxyInfo, node.getUniqueId());
            }

            String nodeClientId = node.getClientId();
            List<LinkTestReport> res = new ArrayList<>();

            String tcpServerScript;
            String linkTestClientScript;
            try {
                tcpServerScript = IOUtils.resourceToString("link_test/thoughput_server_tcp.py", TestLinksJob.class);
                linkTestClientScript = IOUtils.resourceToString("link_test/link_tester.py", TestLinksJob.class);
            } catch (IOException e) {
                throw new JFedException("Failed to get link test script from class resources", e);
            }

            //Start thoughput_server_tcp.py on all other nodes
            List<Pair<Session, Session.Command>> runningThroughputServers = new ArrayList<>();
            try {
                for (FXRspecInterface iface : node.getInterfaces()) {
                    for (FXRspecInterface otherIface : iface.getLink().getInterfaces()) {
                        //noinspection ObjectEquality
                        if (otherIface == iface)
                            continue;
                        FXRspecNode targetNode = otherIface.getNode();

                        LOG.debug("getSSHClient for node " + targetNode.getUniqueId());
                        SSHClient targetSsh = getSSHClient(targetNode, null);

                        if (!targetSsh.isConnected()) {
                            throw new SshException("SSH connection to " + targetNode.getUniqueId() + " is not connected.");
                        }

                        try {
                            LOG.debug("uploading thoughput_server_tcp.py to node " + targetNode.getUniqueId());
                            try (SFTPClient sftpClient = targetSsh.newSFTPClient()) {
                                sftpClient.getFileTransfer().upload(
                                        new StringMemorySourceFile("thoughput_server_tcp.py", tcpServerScript),
                                        "/tmp/");
                            }
                            LOG.debug("finished uploading thoughput_server_tcp.py to node");
                        } catch (IOException e) {
                            throw new SshException("Failed to upload thoughput_server_tcp.py to " + targetNode.getUniqueId(), e);
                        }

                        LimitedLiveLog log = new LimitedLiveLogImpl(10_000);
                        try {
                            Session session = targetSsh.startSession();
                            session.allocateDefaultPTY(); //needed to manually send CTRL-C later.
                            LOG.debug("executing thoughput_server_tcp.py on node " + targetNode.getUniqueId());

                            Session.Command c = session.exec("python /tmp/thoughput_server_tcp.py");
                            runningThroughputServers.add(new Pair<>(session, c));

                            new InputStreamToLogsThread(c.getInputStream(), LOG, Level.DEBUG,
                                    "tts-" + targetNode.getUniqueId() + "-OUTPUT: ", log).start();
                            new InputStreamToLogsThread(c.getErrorStream(), LOG, Level.WARN,
                                    "tts-" + targetNode.getUniqueId() + "-ERROR: ", log).start();
                        } catch (ConnectionException | TransportException e) {
                            throw new SshException("Failed to start thoughput_server_tcp.py on " + targetNode.getUniqueId(), e);
                        }

                    }
                }

                LOG.debug("finished starting thoughput_server_tcp.py");
                //give servers some extra time to start
                Thread.sleep(250);

                Map<String, FXRspecNode> otherNodeByClientId = new HashMap<>();
                for (FXRspecInterface iface : node.getInterfaces()) {
                    String ifaceClientId = iface.getClientId();
                    String command = String.format("python /tmp/link_tester.py %s %s", nodeClientId, ifaceClientId);
                    for (FXRspecInterface otherIface : iface.getLink().getInterfaces()) {
                        //noinspection ObjectEquality
                        if (otherIface == iface)
                            continue;
                        FXRspecNode targetNode = otherIface.getNode();
//                    if (otherIface.getIpAddresses() != null && !otherIface.getIpAddresses().isEmpty()) {
//                        String ip = otherIface.getIpAddresses().get(0).getAddress();
//                        //do we need the target IP?
//                    }

                        String targetNodeId = targetNode.getClientId();

                        otherNodeByClientId.put(targetNodeId, targetNode);

                        command += " " + targetNodeId;
                    }

                    //Execute link test script on node, and process result to make LinkTestReports
                    LOG.debug("Fetching connection to " + node.getUniqueId());
                    SSHClient ssh = getSSHClient(node, null);
                    try {
                        if (!ssh.isConnected()) {
                            throw new SshException("SSH connection to " + node.getUniqueId() + " is not connected.");
                        }

                        try {
                            LOG.debug("uploading link_tester.py to node " + node.getUniqueId());
                            try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                                sftpClient.getFileTransfer().upload(
                                        new StringMemorySourceFile("link_tester.py", linkTestClientScript),
                                        "/tmp/");
                            }
                        } catch (IOException e) {
                            throw new SshException("Failed to upload link_tester.py to " + node.getUniqueId(), e);
                        }

                        List<Map> linkTestRes;
                        StringBuilder stdErrString = new StringBuilder();
//                StringBuilder allOutput = new StringBuilder();
                        LOG.debug("executing link_tester.py on node " + node.getUniqueId());
                        try (Session session = ssh.startSession()) {
                            Session.Command c = session.exec(command);


                            //possible hang: if stderr has lots of output, but stdout has none,
                            //               and the script does not exit or close stdout, the first loop will hang.
                            //However, because we know this script, we know this should never occur.

                            Exception inputStreamEx = null;
                            try (InputStream inputStream = c.getInputStream()) {
//                        BufferedReader stdInput = new BufferedReader(new InputStreamReader(inputStream));
//                        while ((line = stdInput.readLine()) != null) {
//                            allOutput.append(line);
//                        }
                                linkTestRes = MAPPER.readValue(inputStream,
                                        MAPPER.getTypeFactory().constructCollectionType(List.class, Map.class));
                                LOG.debug("processed output of link_tester.py");
                            } catch (Exception e) {
                                inputStreamEx = e;
                                linkTestRes = null;
                                LOG.warn("Delaying exception in link_tester.py inputstream read", e);
                            }

                            try (InputStream errStream = c.getErrorStream()) {
                                BufferedReader stdErr = new BufferedReader(new InputStreamReader(errStream));
                                String line;
                                while ((line = stdErr.readLine()) != null) {
//                            allOutput.append(line);
                                    LOG.warn("link_tester.py on " + node.getUniqueId() + " STDERR: " + line);
                                    stdErrString.append(line).append("\n");
                                }
                            }

                            LOG.debug("waiting for link_tester.py");
                            c.join();
                            LOG.debug("link_tester.py finished with exit status " + c.getExitStatus());
                            if (inputStreamEx != null) {
                                throw inputStreamEx;
                            }
                            if (c.getExitStatus() != 0) {
                                throw new SshException("Exit status " + c.getExitStatus() + " for link_tester.py on " + node.getUniqueId());
                            }

                        } catch (Exception e) {
                            linkTestRes = null;
                            throw new SshException("Failure in link_tester.py on " + node.getUniqueId(), e);
                        }
                        for (Map linkTestResMap : linkTestRes) {
                            String targetClientId = (String) linkTestResMap.get("target");
                            FXRspecNode targetNode = otherNodeByClientId.get(targetClientId);
                            try {
                                LinkTestReport report = LinkTestReport.createFromJson(
                                        node, targetNode, getExpectedLinkSpeed_Mbps(node, targetNode),
                                        linkTestResMap, stdErrString.toString());
                                res.add(report);
                                experiment.getLinkTestLogger().fireReport(report);
                            } catch (Exception e) {
                                throw new RuntimeException("Failed to process linkTestResMap=" + linkTestResMap, e);
                            }
                        }
                    } finally {
//                    try {
//                        ssh.close();
//                    } catch (IOException e) {
//                        LOG.warn("Exception while closing ssh client to "+node.getUniqueId());
//                    }
                    }
                }
            } finally {
                //Stop throughput_server_tcp.py on all other nodes
                for (Pair<Session,Session.Command> p: runningThroughputServers) {
                    Session session = p.getKey();
                    Session.Command command = p.getValue();
                    try {
                        LOG.debug("sending SIGINT to thoughput_server_tcp.py");

//                        command.signal(Signal.INT); //sadly, this doesn't work, even for OpenSSH

                        //manually send CTRL-C
                        try {
                            command.getOutputStream().write(3);
                            command.getOutputStream().flush();
                        } catch (ConnectionException e) {
                            LOG.warn("Ignoring error sending manual CTRL-C", e);
                        }

                        try {
                            command.join(1, TimeUnit.SECONDS); //should not take long
                            int exitStatus = command.getExitStatus();
                            LOG.debug("thoughput_server_tcp.py finished with exitStatus="+exitStatus);
                        } catch (ConnectionException e) {
                            LOG.warn("Ignoring command.join timeout", e);
                        }
                    } catch (IOException e) {
                        LOG.warn("Exception trying to stop thoughput_server_tcp.py -> will be ignored", e);
                    } finally {
                        try {
                            session.close();
                        } catch (TransportException | ConnectionException e) {
                            LOG.warn("Exception in session.close() while trying to stop thoughput_server_tcp.py -> will be ignored", e);
                        }
                    }
                }
            }

            //This is unused, but might be good to know if the error pops up again.
            //It was probably caused by using  command.signal(Signal.INT)   (as it went away when that was replaced)
//            // *** WORKAROUND ***
//            //something very strange is going on when I don't close all SSH connections:
//            //  Got notified of net.schmizz.sshj.transport.TransportException: [PROTOCOL_ERROR] server_input_channel_req: unknown channel 0
//            //possibly a net.schmizz.sshj bug.
//            //set <logger name="net.schmizz" level="debug"/>  in logback.xml to clearly see the problem.
//            closeAllSshConnections();

            return res;
        }

        public Collection<FXRspecNode> getNodesToTest() {
            return nodesToTest;
        }

        public List<LinkTestReport> getLinkReports() {
            return linkReports;
        }
    }


    /**
     * See FEDIBBTDEV-3766
     *
     * Link speed is typically 1Gbit/s, but not always,
     * for example, VM's on virtual wall have only 100Mbit/s
     *
     * The expected link speed is hardcoded in this method for now.
     * If needed, this might later be added to jFed config.
     *
     * @param sourceNode the node on which the bandwidth tester runs
     * @param targetNode the node to which the bandwidth is tested
     * @return the expected link speed between the 2 nodes
     */
    private static int getExpectedLinkSpeed_Mbps(@Nonnull FXRspecNode sourceNode,
                                                 @Nonnull FXRspecNode targetNode) {
        if (Objects.equals(sourceNode.getSliverTypeName(), "emulab-xen") ||
                Objects.equals(targetNode.getSliverTypeName(), "emulab-xen")) {
            return 100;
        }
        if (Objects.equals(sourceNode.getSliverTypeName(), "emulab-openvz") ||
                Objects.equals(targetNode.getSliverTypeName(), "emulab-openvz")) {
            return 100;
        }
        return 1000;
    }
}
